@@ -30,7 +30,9 @@ import time
 from .failmanager import FailManagerEmpty
 from .filter import FileFilter
 from .mytime import MyTime
-from ..helpers import getLogger
+from .utils import Utils
+from ..helpers import getLogger, logging
+
 
 # Gets the instance of the logger.
 logSys = getLogger(__name__)
@@ -77,6 +79,15 @@ class FilterPoll(FileFilter):
 		del self.__prevStats[path]
 		del self.__file404Cnt[path]
 
+	##
+	# Get a modified log path at once
+	#
+	def getModified(self, modlst):
+		for filename in self.getLogPaths():
+			if self.isModified(filename):
+				modlst.append(filename)
+		return modlst
+
 	##
 	# Main loop.
 	#
@@ -86,17 +97,25 @@ class FilterPoll(FileFilter):
 
 	def run(self):
 		while self.active:
-			if logSys.getEffectiveLevel() <= 6:
-				logSys.log(6, "Woke up idle=%s with %d files monitored",
-						   self.idle, len(self.getLogs()))
-			if not self.idle:
+			try:
+				if logSys.getEffectiveLevel() <= 6:
+					logSys.log(6, "Woke up idle=%s with %d files monitored",
+							   self.idle, self.getLogCount())
+				if self.idle:
+					if not Utils.wait_for(lambda: not self.active or not self.idle, 
+						self.sleeptime * 10, self.sleeptime
+					):
+						self.ticks += 1
+						continue
 				# Get file modification
-				for container in self.getLogs():
-					filename = container.getFileName()
-					if self.isModified(filename):
-						self.getFailures(filename)
-						self.__modified = True
+				modlst = []
+				Utils.wait_for(lambda: not self.active or self.getModified(modlst),
+					self.sleeptime)
+				for filename in modlst:
+					self.getFailures(filename)
+					self.__modified = True
 
+				self.ticks += 1
 				if self.__modified:
 					try:
 						while True:
@@ -104,14 +123,15 @@ class FilterPoll(FileFilter):
 							self.jail.putFailTicket(ticket)
 					except FailManagerEmpty:
 						self.failManager.cleanup(MyTime.time())
-					self.dateDetector.sortTemplate()
 					self.__modified = False
-				time.sleep(self.sleeptime)
-			else:
-				time.sleep(self.sleeptime)
-		logSys.debug(
-			(self.jail is not None and self.jail.name or "jailless") +
-					 " filter terminated")
+			except Exception as e: # pragma: no cover
+				if not self.active: # if not active - error by stop...
+					break
+				logSys.error("Caught unhandled exception in main cycle: %r", e,
+					exc_info=logSys.getEffectiveLevel()<=logging.DEBUG)
+				# incr common error counter:
+				self.commonError()
+		logSys.debug("[%s] filter terminated", self.jailName)
 		return True
 
 	##
@@ -124,29 +144,41 @@ class FilterPoll(FileFilter):
 		try:
 			logStats = os.stat(filename)
 			stats = logStats.st_mtime, logStats.st_ino, logStats.st_size
-			pstats = self.__prevStats[filename]
-			self.__file404Cnt[filename] = 0
-			if logSys.getEffectiveLevel() <= 7:
+			pstats = self.__prevStats.get(filename, (0))
+			if logSys.getEffectiveLevel() <= 5:
 				# we do not want to waste time on strftime etc if not necessary
 				dt = logStats.st_mtime - pstats[0]
-				logSys.log(7, "Checking %s for being modified. Previous/current stats: %s / %s. dt: %s",
+				logSys.log(5, "Checking %s for being modified. Previous/current stats: %s / %s. dt: %s",
 				           filename, pstats, stats, dt)
 				# os.system("stat %s | grep Modify" % filename)
+			self.__file404Cnt[filename] = 0
 			if pstats == stats:
 				return False
-			else:
-				logSys.debug("%s has been modified", filename)
-				self.__prevStats[filename] = stats
-				return True
-		except OSError, e:
-			logSys.error("Unable to get stat on %s because of: %s"
-						 % (filename, e))
+			logSys.debug("%s has been modified", filename)
+			self.__prevStats[filename] = stats
+			return True
+		except Exception as e:
+			# stil alive (may be deleted because multi-threaded):
+			if not self.getLog(filename) or self.__prevStats.get(filename) is None:
+				logSys.warning("Log %r seems to be down: %s", filename, e)
+				return False
+			# log error:
+			if self.__file404Cnt[filename] < 2:
+				if e.errno == 2:
+					logSys.debug("Log absence detected (possibly rotation) for %s, reason: %s",
+							 filename, e)
+				else: # pragma: no cover
+					logSys.error("Unable to get stat on %s because of: %s",
+							 filename, e, 
+							 exc_info=logSys.getEffectiveLevel()<=logging.DEBUG)
+			# increase file and common error counters:
 			self.__file404Cnt[filename] += 1
-			if self.__file404Cnt[filename] > 2:
-				logSys.warning("Too many errors. Setting the jail idle")
-				if self.jail is not None:
-					self.jail.idle = True
-				else:
-					logSys.warning("No jail is assigned to %s" % self)
+			self.commonError()
+			if self.__file404Cnt[filename] > 50:
+				logSys.warning("Too many errors. Remove file %r from monitoring process", filename)
 				self.__file404Cnt[filename] = 0
+				self.delLogPath(filename)
 			return False
+
+	def getPendingPaths(self):
+		return self.__file404Cnt.keys()
